<div class="breadcrumbs">
{%- for node in schema.nodes_from_root -%}
    {%- if loop.first -%}
        root
    {%- else -%}
        <a href="#{{ node.html_id }}">{{ node.name_for_breadcrumbs }}</a>
    {%- endif -%}
    {%- if not loop.last %}
        <svg width="1em" height="1em" viewBox="0 0 16 16" class="bi bi-arrow-right-short" fill="currentColor" xmlns="http://www.w3.org/2000/svg">
            <path
                fill-rule="evenodd"
                d="M4 8a.5.5 0 0 1 .5-.5h5.793L8.146 5.354a.5.5 0 1 1 .708-.708l3 3a.5.5 0 0 1 0 .708l-3 3a.5.5 0 0 1-.708-.708L10.293 8.5H4.5A.5.5 0 0 1 4 8z"
            />
        </svg>
    {% endif -%}
{%- endfor -%}
</div>